(ns puppetlabs.rbac.shiro-utils
  (:import [java.util Hashtable]
           [javax.naming Context]
           [javax.naming.ldap InitialLdapContext]
           [com.jolbox.bonecp BoneCPDataSource]
           [org.apache.shiro                 SecurityUtils]
           [org.apache.shiro.authc           UsernamePasswordToken IncorrectCredentialsException]
           [org.apache.shiro.realm.jdbc      JdbcRealm]
           [org.apache.shiro.realm.ldap      JndiLdapRealm JndiLdapContextFactory]
           [org.apache.shiro.session.mgt     DefaultSessionKey]
           [org.apache.shiro.subject         SimplePrincipalCollection Subject$Builder]
           [org.apache.shiro.web.mgt         DefaultWebSecurityManager ]
           [org.apache.shiro.web.servlet     ShiroHttpServletRequest]
           [org.apache.shiro.web.session.mgt DefaultWebSessionManager]))


(def session-key-attr
  "This string is used with getAttribute to find the principal(s)
  associated with a session."
  "org.apache.shiro.subject.support.DefaultSubjectContext_PRINCIPALS_SESSION_KEY")

(defn create-security-manager [realms session-timeout]
  (let [sec-mgr  (DefaultWebSecurityManager. realms)
        sess-mgr (doto (DefaultWebSessionManager.)
                   (.setGlobalSessionTimeout session-timeout))]
    ;; TODO this may not be desirable at some point; we can add a
    ;; session-cook-path config option to change it.
    (.setPath (.getSessionIdCookie sess-mgr) "/")

    (doto sec-mgr
      (.setSessionManager sess-mgr))))

(defn login->subject
  "Given a Shiro security manager and a subject's login, returns a
  DelegatingSubject that can be used for authn/authz queries. Note
  that it is not a WebSubject and will have no notion of session."
  [sec-mgr subject-login]
  (if (nil? subject-login)
    nil
    (.buildSubject
     (.principals
      (Subject$Builder. sec-mgr) (SimplePrincipalCollection. subject-login "rbac")))))

(defn session-id->subject
  "Given a session ID, return a (non-web) Subject useful only for
  permission checking. Returns nil when no subject is found for the
  session id."
  [sec-mgr session-id]
  (try
    (let [session-key (DefaultSessionKey. session-id)
          session     (.getSession sec-mgr session-key)
          principals  (.getAttribute session session-key-attr)]
      (.buildSubject
       (.principals
        (Subject$Builder. sec-mgr) principals)))
    (catch org.apache.shiro.session.UnknownSessionException e
      nil)))

(defn is-permitted?
  "Given a Shiro Subject, check to see if it has the given permission
  string."
  [subject perm-str]
  (if (nil? subject)
    false
    (.isPermitted subject perm-str)))

(defn are-permitted?
  "Given a Shiro subject and a list of permission strings, returns a
  vector of booleans that map to the original permissions."
  [subject perm-strs]
  (if (nil? subject)
    (vec (repeat (count perm-strs) false))
    (if (empty? perm-strs)
      []
      (vec (.isPermitted subject (into-array perm-strs))))))

(defn authenticated? [subject]
  (.isAuthenticated subject))

;(def current-subject-authenticated?
;  (comp authenticated? current-subject))

(defn login!
  [subject user pw]
  (let [tok (UsernamePasswordToken. user pw)
        subj subject]
    (.login subj tok)))

(defn logout! [subj]
    (.logout subj))

(defn get-data-source [db]
  (Class/forName "org.postgresql.Driver")
  (let [ds (BoneCPDataSource.)]
    (doto ds
      (.setJdbcUrl (str "jdbc:"
                        (:subprotocol db) ":"
                        (:subname db)))
      (.setUsername (:user db))
      (.setPassword (:password db)))
    ds))

(def user-roles-query
  "SELECT DISTINCT r.display_name
   FROM subjects s
   LEFT JOIN groupings g ON s.id = g.ruid
   JOIN subject_roles sr ON (s.id = sr.sid OR g.gid = sr.sid)
   JOIN roles r ON sr.rid = r.id
   WHERE s.login = ?")

(def permissions-query
  "SELECT p.object_type || ':' || p.action || ':' || p.instance
          AS permission
   FROM permissions p JOIN
        roles_permissions rp ON rp.pid = p.id JOIN
        roles r ON r.id = rp.rid
   WHERE r.display_name = ?")

(def authentication-query
  "SELECT password FROM subjects WHERE login = ?")

(defn create-jdbc-realm [db-conn-pool]
  (let [realm (proxy [JdbcRealm] [])]
    (doto realm
      (.setCredentialsMatcher (doto (org.apache.shiro.authc.credential.PasswordMatcher.)
                                (.setPasswordService (org.apache.shiro.authc.credential.DefaultPasswordService.))))
      (.setDataSource db-conn-pool)
      (.setPermissionsLookupEnabled true)
      (.setUserRolesQuery user-roles-query)
      (.setPermissionsQuery permissions-query)
      (.setAuthenticationQuery authentication-query))))

(defn user-dn-template [{:keys [user_rdn base_dn user_lookup_attr]}]
  (format "%s={0},%s,%s" user_lookup_attr user_rdn base_dn))

(defn build-ldap-ctx-factory [ssl]
  (if ssl
    (proxy [JndiLdapContextFactory] []
      (createLdapContext [env]
        ;; Disable cert validation. We'll actually validate certs in 3.5.
        (.put env "java.naming.ldap.factory.socket" "com.puppetlabs.rbac.DummySSLSocketFactory")
        (proxy-super createLdapContext env)))
    (JndiLdapContextFactory.)))

(defn connect-realm [{:keys [hostname port login password ssl] :as settings} realm]
  (let [url          (format "ldap%s://%s:%s" (if ssl "s" "") hostname port)
        user-dn-tmpl (user-dn-template settings)
        ctx-factory  (build-ldap-ctx-factory ssl)]

    (doto ctx-factory
      (.setUrl url)
      (.setSystemUsername login)
      (.setSystemPassword password))

    (.setContextFactory realm ctx-factory)

    ;; TODO might need setAuthenticationMechanism at some point if
    ;; "simple" authn ceases to suffice.
    (doto realm (.setUserDnTemplate user-dn-tmpl))))

(defn create-ds-realm [{:keys [ssl] :as settings}]
    (connect-realm settings (JndiLdapRealm.)))

(defn password-matches?
  "Compare the provided password with the current password for
   `login`. Note that hashes are salted so the same password hashed
   two times will not be equal, using the Shiro methods for doing the
   match takes the salts into account"
  [realm login password]
  (try
    (let [auth-token (UsernamePasswordToken. login password)]
      (-> realm
          .getCredentialsMatcher
          (.doCredentialsMatch auth-token (.getAuthenticationInfo realm auth-token))))
    (catch IncorrectCredentialsException _
      false)))
